import re
from dataclasses import dataclass
from datetime import datetime


@dataclass
class FirmwareVersion:
    """Represents a firmware version, parsing all components used by Lobaro firmware.

    Raises:
        ValueError: If the provided version string does not match the expected format.
    """

    version_string: str
    """The complete version string to be parsed."""

    major: int = 0
    """Major version number."""

    minor: int = 0
    """Minor version number."""

    patch: int = 0
    """Patch version number."""

    commits: int = 0
    """Number of commits since the last version tag."""

    commit: str | None = None
    """Commit hash associated with this version."""

    dirty: bool = False
    """Indicates if there are uncommitted changes in the source."""

    unknown: bool = False
    """Flag for any unknown additional information in the version string."""

    _VERSION_REGEX = re.compile(
        r"(?P<major>\d+)\.(?P<minor>\d+)\.(?P<patch>\d+)"
        r"(?:-(?P<commits>\d+)-g(?P<commit>[0-9a-f]+))?"
        r"(?P<dirty>-dirty)?(?P<unknown>-unknown)?"
    )
    """Precompiled regex for parsing all parts of a version string."""

    def __post_init__(self):
        """Parses the version string and updates the object's attributes accordingly.

        This method is automatically called after the data class has been initialized.
        """
        m = self._VERSION_REGEX.match(self.version_string)
        try:
            self.major = int(m.group("major"))
            self.minor = int(m.group("minor"))
            self.patch = int(m.group("patch"))
            self.commits = int(m.group("commits") or 0)
            self.commit = m.group("commit")
            self.dirty = bool(m.group("dirty"))
            self.unknown = bool(m.group("unknown"))
        except (AttributeError, ValueError, TypeError, IndexError, KeyError) as exc:
            raise ValueError(
                f"Invalid version string: '{self.version_string}'"
            ) from exc


@dataclass
class FirmwareID:
    """Represents a firmware id, parsing all components used by Lobaro firmware.

    Raises:
        ValueError: If the provided id does not match the expected format.
    """

    id_string: str
    """The complete firmware identifier string to be parsed."""

    name: str | None = None
    """The name of the firmware."""

    version: FirmwareVersion | None = None
    """The version of the firmware."""

    variants: list[str] | None = None
    """A list of all variants of the firmware, e.g. ['hw3', 'alt_phy']."""

    built: str | None = None
    """The build date of firmware, e.g. 'Jan 01 2021 00:00:00'"""

    _IDENTIFIER_RE = re.compile(
        # Name group, non-greedy match up to the first space
        r"^(?P<name>.+?)\s+"
        # Version group, matches a semantic versioning pattern
        r"v(?P<version>[0-9]+(?:\.[0-9]+){2}(?:-[\w]+)?(?:-\d+-g[0-9a-f]+)?)"
        # Optional variant group, matches anything after a '+' until a space or end
        r"(?:\+(?P<variant>[^\s]+))?"
        # Optional additional group,
        # matches before parentheses if not directly next to variant
        r"(?:\s+(?P<additional>[^\(\)]+?))?"
        # Date group, matches everything inside the parentheses, made entirely optional
        r"(?:\s*\((?P<date>.+?)\))?$",
    )
    _IDENTIFIER_RE_ORIGIN = re.compile(
        # Name group, non-greedy match up to the first +
        r"^(?P<name>.+?)\+"
        # Version group, matches a semantic versioning pattern
        r"(?P<version>[0-9]+(?:\.[0-9]+){2}(?:-[\w]+)?(?:-\d+-g[0-9a-f]+)?)"
        # Optional variant group, matches anything after a '+' until a space or end
        r"(?:\+(?P<variant>[^\s]+))?"
        # Optional additional group,
        # matches before parentheses if not directly next to variant
        r"(?:\s+(?P<additional>[^\(\)]+?))?"
        # Date group, matches everything inside the parentheses, made entirely optional
        r"(?:\s*\((?P<date>.+?)\))?$",
    )
    """Precompiled regex for parsing all parts of a firmware identifier string."""

    def __post_init__(self):
        """Parses the firmware identifier string and updates attributes accordingly."""
        try:
            m = self._IDENTIFIER_RE.match(self.id_string)
            if not m:
                m = self._IDENTIFIER_RE_ORIGIN.match(self.id_string)
            groups = m.groupdict()
            self.name = groups["name"]
            self.version = FirmwareVersion(groups["version"])
            self.variants = groups["variant"].split(".") if groups["variant"] else []
            built = groups.get("date", None)
            if built:
                self.built = datetime.strptime(built, "%b %d %Y %H:%M:%S").isoformat()
        except (AttributeError, ValueError, TypeError, IndexError, KeyError) as exc:
            raise ValueError(
                f"Invalid firmware identifier string: '{self.id_string}'"
            ) from exc



import pytest

from lob_hlpr import FirmwareVersion


def test_valid_version():
    """Test with a valid version string."""
    version = FirmwareVersion("1.2.3-4-gabc123-dirty")
    assert version.major == 1
    assert version.minor == 2
    assert version.patch == 3
    assert version.commits == 4
    assert version.commit == "abc123"
    assert version.dirty is True
    assert version.unknown is False


def test_valid_version_with_unknown():
    """Test with a version string that includes 'unknown'."""
    version = FirmwareVersion("1.2.3-unknown")
    assert version.major == 1
    assert version.minor == 2
    assert version.patch == 3
    assert version.commits == 0
    assert version.commit is None
    assert version.dirty is False
    assert version.unknown is True


def test_version_without_optional_parts():
    """Test with a version string that does not include optional parts."""
    version = FirmwareVersion("1.2.3")
    assert version.major == 1
    assert version.minor == 2
    assert version.patch == 3
    assert version.commits == 0
    assert version.commit is None
    assert version.dirty is False
    assert version.unknown is False


def test_invalid_version():
    """Test with an invalid version string."""
    with pytest.raises(ValueError):
        FirmwareVersion("invalid.version.string")


def test_empty_version():
    """Test with an empty version string."""
    with pytest.raises(ValueError):
        FirmwareVersion("")


def test_partial_version():
    """Test with a partial version string."""
    # Assuming partial versions are not allowed and should raise an error
    with pytest.raises(ValueError):
        FirmwareVersion("1.2")



"""The helpers: lob_hlpr module.

Simple python based helpers for lobaro tools.
"""

from lob_hlpr.hlpr import LobHlpr
from lob_hlpr.lib_types import FirmwareID, FirmwareVersion

__all__ = [
    "LobHlpr",
    "FirmwareID",
    "FirmwareVersion",
]



import binascii
import logging
import logging.handlers
import os
import re
from dataclasses import asdict
from datetime import datetime

from lob_hlpr.lib_types import FirmwareID


def enable_windows_ansi_support():  # pragma: no cover
    """Try to enable ANSI escape sequence support on Windows.

    Works on Windows 10+.
    """
    if os.name != "nt":
        return True  # Non-Windows always supports ANSI

    try:
        import ctypes

        kernel32 = ctypes.windll.kernel32

        # Enable Virtual Terminal Processing
        handle = kernel32.GetStdHandle(-11)  # STD_OUTPUT_HANDLE = -11
        mode = ctypes.c_uint32()
        if kernel32.GetConsoleMode(handle, ctypes.byref(mode)):
            new_mode = (
                mode.value | 0x0004
            )  # ENABLE_VIRTUAL_TERMINAL_PROCESSING = 0x0004
            if kernel32.SetConsoleMode(handle, new_mode):
                return True
    except Exception:
        pass

    return False


# Determine if ANSI colors will work
_USE_COLOR = enable_windows_ansi_support()


class LobHlpr:
    """Helper functions for Lobaro tools."""

    @staticmethod
    def sn_vid_pid_to_regex(
        sn: str | None = None, vid: str | None = None, pid: str | None = None
    ):
        r"""Convert serial number, VID and PID to regex.

        This is a convenience function for serial.tools.list_ports.grep.
        Some examples of output that one would get is:
        port: /dev/ttyUSB0,
        desc: CP2102 USB to UART Bridge Controller -
            CP2102 USB to UART Bridge Controller,
        hwid: USB VID:PID=10C4:EA60 SER=KW-0001 LOCATION=1-1

        Args:
            sn (str): The serial number to search for.
            vid (str): The VID to search for.
            pid (str): The PID to search for.

        Returns:
            str: The regex string.

        Examples:
            >>> tst = LobHlpr.sn_vid_pid_to_regex
            >>> print(tst(sn="KW-0001", vid="10C4", pid="EA60"))
            VID:PID=10C4:EA60.+SER=KW\-0001
            >>> print(tst(sn="KW-0001"))
            VID:PID=.*:.*.+SER=KW\-0001
            >>> print(tst(pid="EA60", vid="10C4"))
            VID:PID=10C4:EA60.+SER=.*
            >>> print(tst(sn="KW-0001", vid="10C4"))
            VID:PID=10C4:.*.+SER=KW\-0001
            >>> print(tst(sn="KW-0001", pid="EA60"))
            VID:PID=.*:EA60.+SER=KW\-0001
            >>> print(tst(sn="SPECIAL-.*"))
            VID:PID=.*:.*.+SER=SPECIAL\-\.\*
        """
        if sn is None:
            sn = ".*"
        else:
            sn = re.escape(sn)
        return f"VID:PID={vid or '.*'}:{pid or '.*'}.+SER={sn}"

    @staticmethod
    def _print_color(*args, color=None, **kwargs):
        """Print with color if supported."""
        # ANSI color codes
        RESET = "\033[0m"
        RED = "\033[31m"
        GREEN = "\033[32m"
        YELLOW = "\033[33m"
        if color is None or not _USE_COLOR:
            print(*args, flush=True, **kwargs)
            return
        text = kwargs.get("sep", " ").join(str(a) for a in args)
        color = color.lower()
        code = None
        if "red" == color:
            code = RED
        elif "green" == color:
            code = GREEN
        elif "yellow" == color:
            code = YELLOW
        if code is not None:
            print(f"{code}{text}{RESET}", flush=True, **kwargs)
        else:
            print(text, flush=True, **kwargs)

    @staticmethod
    def lob_print(log_path: str, *args, **kwargs):
        """Print to the console and log to a file.

        The log file is rotated when it reaches 256MB and the last two
        log files are kept. This can write all log messages to the log file
        only if the log handlers are set (i.e. basicConfig loglevel is Debug).

        Args:
            log_path: The path to the log file.
            *args: Arguments to print.
            **kwargs: Additional keyword arguments.
                color (str, optional): If provided, prints the message in color
                    to the console. Supported values are "red", "green", and "yellow".
                    If colors are not supported by the terminal, output will be
                    uncolored.
        """
        color = kwargs.pop("color", None)
        LobHlpr._print_color(*args, color=color, **kwargs)

        # get the directory from the log_path
        log_dir = os.path.dirname(log_path)
        os.makedirs(log_dir, exist_ok=True)
        logger = logging.getLogger("lob_hlpr")
        # Check to see if the file handler was already set up for root logger
        logger.propagate = False  # Prevent propagation to root logger
        logger.setLevel(logging.INFO)
        root_logger = logging.getLogger()

        # Check if our file handler is already attached to root logger
        has_file_handler = any(
            isinstance(h, logging.handlers.RotatingFileHandler)
            and h.baseFilename == os.path.abspath(log_path)
            for h in root_logger.handlers
        )

        if not has_file_handler:
            ch = logging.handlers.RotatingFileHandler(
                log_path, maxBytes=268435456, backupCount=2
            )

            formatter = logging.Formatter(
                "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
            )

            ch.setFormatter(formatter)

            # Add handler to root logger so all loggers inherit it
            root_logger.addHandler(ch)
            logger.addHandler(ch)

        logger.info(*args, **kwargs)

    @staticmethod
    def ascleandict(dclass, remove_false=False):
        """Convert a dataclass to a dictionary and remove None values.

        Largely generated by AI...

        Args:
            dclass: The dataclass instance to convert.
            remove_false: If True, also remove boolean fields that are False.

        Returns:
            dict: The cleaned dictionary without empty values.
        """

        def clean_value(v):
            """Recursively clean nested structures."""
            if isinstance(v, dict):
                cleaned = {
                    k: clean_value(val)
                    for (k, val) in v.items()
                    if (val is not None)
                    and not (isinstance(val, list) and len(val) == 0)
                    and not (isinstance(val, dict) and len(val) == 0)
                    and not (remove_false and (isinstance(val, bool) and val is False))
                }
                # Keep removing empty dicts/lists until nothing changes
                while True:
                    filtered = {
                        k: v
                        for (k, v) in cleaned.items()
                        if not (isinstance(v, dict) and len(v) == 0)
                        and not (isinstance(v, list) and len(v) == 0)
                    }
                    if len(filtered) == len(cleaned):
                        break
                    cleaned = filtered
                return cleaned
            elif isinstance(v, list):
                cleaned_items = [clean_value(item) for item in v]
                # Filter out empty dicts and lists from the result
                return [
                    item
                    for item in cleaned_items
                    if not (isinstance(item, dict) and len(item) == 0)
                    and not (isinstance(item, list) and len(item) == 0)
                ]
            return v

        result = asdict(
            dclass,
            dict_factory=lambda x: {
                k: v
                for (k, v) in x
                if (v is not None)
                and not (isinstance(v, list) and len(v) == 0)
                and not (isinstance(v, dict) and len(v) == 0)
                and not (remove_false and (isinstance(v, bool) and v is False))
            },
        )
        return clean_value(result)

    @staticmethod
    def unix_timestamp() -> int:
        """Unix timestamp - milliseconds since 1970.

        Example: 1732266521241
        """
        return int(datetime.now().timestamp() * 1000)

    @staticmethod
    def format_unix_timestamp(timestamp) -> str:
        """Formatted unix timestamp.

        Example: 2024-11-22_10-18-24
        """
        return datetime.fromtimestamp(timestamp / 1000).strftime("%Y-%m-%d_%H-%M-%S")

    @staticmethod
    def parse_dmc(dmc):
        """Parse the dmc scanned by the barcode scanner for the PCB.

        Example input from scanner MPP-OR019504_1-00781 or MPP-M0011554-OR019504_1-00781

        Args:
            dmc (str): The scanned dmc, digital manufacturer code.

        Returns:
            Tuple[str, str, str, str]: The erp_prod_number, batch_number,
                pcba_serial_number, article_number.
        """
        article_number = None
        chunks: list[str] = dmc.split("-")
        if len(chunks) == 4:
            # This: MPP-M0011554-OR019504_1-00781
            article_number = chunks.pop(1)
            # Changes to MPP-OR019504_1-00781
        # We want the OR019504 part
        erp_prod_number = chunks[0] + "-" + re.split("_", chunks[1])[0]

        # Now we want the 1 part
        batch_number = re.split("_", chunks[1])[1]

        # Now we want the 00781 part
        pcba_serial_number = chunks[2]

        return erp_prod_number, batch_number, pcba_serial_number, article_number

    @staticmethod
    def extract_identifier_from_hexfile(hex_str: str):
        """Extract the identifier from a hex file.

        Args:
            hex_str (str): The hex file to search in.

        Returns:
            list: The identifiers found in the hex file.
        """
        r = re.compile(r"^:([0-9a-fA-F]{10,})")
        segments = []
        segment = bytearray()
        segment_address = None
        for idx, line in enumerate(hex_str.split("\n")):
            line = line.replace("\r", "")
            if line == "":
                continue
            m = r.match(line)
            if not m:
                raise ValueError(f"Invalid line {idx} in hexfile: {line}")
            b = binascii.unhexlify(m.group(1))
            if len(b) != b[0] + 5:
                raise ValueError(f"Invalid line in hexfile: {line}")
            addr = int.from_bytes(b[1:3], byteorder="big")
            rec_type = b[3]
            data = b[4:-1]
            if rec_type == 0x04:
                # Extended address
                # (higher 2 bytes of address for following data records)
                extended_address = int.from_bytes(data, byteorder="big")
            elif rec_type == 0x00:
                # Data record
                full_address = (extended_address << 16) | addr
                if segment_address is None:
                    segment_address = full_address
                    segment = bytearray(data)
                elif full_address == segment_address + len(segment):
                    segment.extend(data)
                else:
                    segments.append((segment_address, segment))
                    segment = bytearray(data)
                    segment_address = full_address
            elif rec_type == 0x01:
                # End of file
                segments.append((segment_address, segment))

        identifiers = []
        hexinfo_regex = re.compile(b">==HEXINFO==>(.*?)<==HEXINFO==<")
        for seg in segments:
            mat = hexinfo_regex.search(seg[1])
            if mat:
                identifiers.append(mat.groups()[0].decode())
        if len(identifiers) == 0:
            raise ValueError("No firmware identifier found in hexfile")
        return identifiers

    @staticmethod
    def fw_id_from_fw_file(fw_file: str, contains: str | None = None) -> FirmwareID:
        """Extract the firmware identifier from a firmware file.

        Args:
            fw_file (str): The path to the firmware file.
            contains (str): Optional filter to make sure result contains.

        Returns:
            FirmwareID: The firmware identifier.

        Raises:
            ValueError: If no or too many firmware identifier is found in the file.
        """
        with open(fw_file) as f:
            hex_str = f.read()
        identifiers = LobHlpr.extract_identifier_from_hexfile(hex_str)
        if contains:
            identifiers = [i for i in identifiers if contains in i]
        if not identifiers:
            raise ValueError(f"No firmware identifier found in {fw_file}")
        if len(identifiers) > 1:
            raise ValueError(
                f"Multiple firmware identifiers found in {fw_file}: {identifiers}"
            )
        return FirmwareID(identifiers[0])



import logging
from dataclasses import dataclass

import pytest

from lob_hlpr import LobHlpr as hlp


def test_parse_dmc_pass():
    """Test passing branches of the parse_dmc function."""
    assert hlp.parse_dmc("MPP-OR023282_1-00002") == (
        "MPP-OR023282",
        "1",
        "00002",
        None,
    )
    assert hlp.parse_dmc("MPP-M0011554-OR023282_1-00002") == (
        "MPP-OR023282",
        "1",
        "00002",
        "M0011554",
    )


def test_parse_dmc_fail():
    """Test failing branches of the parse_dmc function."""
    with pytest.raises(IndexError):
        hlp.parse_dmc("MPP-OR023282_1")


# A bit of looking through a hex file and manual parsing allows me to provide
# the following data for testing.
HEX_STRINGS = [
    """:020000040000FA

:100400003E3D3D484558494E464F3D3D3E6170702A
:100410002D6D6375626F6F742D6E72663931363073
:100420002D7365632076312E332E302B4D422054B0
:100430005A32202853657020203920323032312042
:1004400031323A32383A3533293C3D3D48455849F6
:070450004E464F3D3D3C000C
:00000001FF""",
    """:020000040000FA
:10C800003E3D3D484558494E464F3D3D3E61707066
:10C810002D626F6F742D6E7266393136302D7365EF
:10C82000632076312E382E3220545A322028466525
:10C830006220203820323032332031373A30333AD8
:10C840003337293C3D3D484558494E464F3D3D3CD8
:00000001FF""",
    """:020000040000FA
:100800003E3D3D484558494E464F3D3D3E61707026
:100810002D6E7266393136302D776D6275732076A4
:10082000302E32332E362B68773320545A3220281C
:100830004A616E20323720323032342031343A323D
:10084000303A3436293C3D3D484558494E464F3DA7
:100850003D3C7365637572653D312E382E303B70BB
:10086000696E3D4D4200FFFFFFFFFFFFFFFFFFFFEF
:00000001FF""",
    """:020000040000FA
:100400003E3D3D484558494E464F3D3D3E6170702A
:100410002D6D6375626F6F742D6E72663931363073
:100420002D7365632076312E332E302B4D422054B0
:100430005A32202853657020203920323032312042
:1004400031323A32383A3533293C3D3D48455849F6
:070450004E464F3D3D3C000C
:10C800003E3D3D484558494E464F3D3D3E61707066
:10C810002D626F6F742D6E7266393136302D7365EF
:10C82000632076312E382E3220545A322028466525
:10C830006220203820323032332031373A30333AD8
:10C840003337293C3D3D484558494E464F3D3D3CD8
:020000040000FA
:100800003E3D3D484558494E464F3D3D3E61707026
:100810002D6E7266393136302D776D6275732076A4
:10082000302E32332E362B68773320545A3220281C
:100830004A616E20323720323032342031343A323D
:10084000303A3436293C3D3D484558494E464F3DA7
:100850003D3C7365637572653D312E382E303B70BB
:10086000696E3D4D4200FFFFFFFFFFFFFFFFFFFFEF
:020000021200EA
:00000001FF""",
]


@pytest.mark.parametrize(
    "hex_str,expected",
    [
        (
            HEX_STRINGS[0],
            ["app-mcuboot-nrf9160-sec v1.3.0+MB TZ2 (Sep  9 2021 12:28:53)"],
        ),
        (HEX_STRINGS[1], ["app-boot-nrf9160-sec v1.8.2 TZ2 (Feb  8 2023 17:03:37)"]),
        (HEX_STRINGS[2], ["app-nrf9160-wmbus v0.23.6+hw3 TZ2 (Jan 27 2024 14:20:46)"]),
        (
            HEX_STRINGS[3],
            [
                "app-mcuboot-nrf9160-sec v1.3.0+MB TZ2 (Sep  9 2021 12:28:53)",
                "app-boot-nrf9160-sec v1.8.2 TZ2 (Feb  8 2023 17:03:37)",
                "app-nrf9160-wmbus v0.23.6+hw3 TZ2 (Jan 27 2024 14:20:46)",
            ],
        ),
    ],
)
def test_extract_identifier_valid_hex(hex_str, expected):
    """Test extract_identifier_from_hexfile with valid HEXINFO."""
    identifier = hlp.extract_identifier_from_hexfile(hex_str)
    print(identifier)
    for exp in expected:
        assert exp in identifier


def test_extract_identifier_invalid_hex():
    """Test extract_identifier_from_hexfile with invalid HEXINFO."""
    # Hex string with invalid HEXINFO
    hex_str = "garbage data"
    with pytest.raises(ValueError):
        hlp.extract_identifier_from_hexfile(hex_str)

    hex_str = ":0102030405"
    with pytest.raises(ValueError):
        hlp.extract_identifier_from_hexfile(hex_str)


def test_extract_identifier_no_hexinfo():
    """Test extract_identifier_from_hexfile with no HEXINFO."""
    # Hex string without HEXINFO, just for structure demonstration
    hex_str = """:020000040000FA
:10C000000102030405060708090A0B0C0D0E0F1011
:00000001FF"""
    with pytest.raises(ValueError):
        hlp.extract_identifier_from_hexfile(hex_str)


def test_log_print_passes(tmp_path, capsys):
    """Test log_print function with valid inputs."""
    test_logger = logging.getLogger("test_logger")
    test_logger.setLevel("INFO")
    test_logger.info("Will not show up in logs")
    test_file = tmp_path / "test.log"
    hlp.lob_print(str(test_file), "Test message")
    hlp.lob_print(str(test_file), "Another test message")
    hlp.lob_print(str(test_file), "red message", color="red")
    hlp.lob_print(str(test_file), "yellow message", color="yellow")
    hlp.lob_print(str(test_file), "green message", color="green")
    hlp.lob_print(str(test_file), "normal message", color="doesn't matter")
    test_logger.info("Will show in logs after lob_print")
    captured = capsys.readouterr()
    assert "Test message" in captured.out
    # Check that only one "Another test message" is in the output
    assert captured.out.count("Another test message") == 1
    assert test_file.exists()
    assert captured.out.count("red message") == 1
    with open(test_file) as f:
        log_content = f.read()
        assert "Test message" in log_content
        assert "Will show in logs after lob_print" in log_content
        assert "Will not show up in logs" not in log_content
        # Check that only one "Another test message" is in the output
        assert log_content.count("Another test message") == 1


def test_as_clean_dict_passes():
    """Test as_clean_dict function with valid inputs."""

    @dataclass
    class MoreNestedTestClass:
        listkey: list | None = None
        dictkey: dict | None = None

    @dataclass
    class NestedTestClass:
        more_nested: MoreNestedTestClass | None = None

    @dataclass
    class TestClass:
        strkey: str | None = None
        intkey: int | None = None
        boolkey: bool | None = None
        listkey: list | None = None
        dictkey: dict | None = None
        nested: NestedTestClass | None = None

    dclass = TestClass()

    assert hlp.ascleandict(dclass) == {}

    dclass.strkey = ""
    dclass.intkey = 0
    dclass.listkey = []
    dclass.dictkey = {}
    dclass.boolkey = False
    dclass.nested = NestedTestClass()
    dclass.nested.more_nested = MoreNestedTestClass(dictkey={}, listkey=[])
    assert hlp.ascleandict(dclass) == {"intkey": 0, "strkey": "", "boolkey": False}
    assert "nested" not in hlp.ascleandict(dclass)

    assert hlp.ascleandict(dclass, remove_false=True) == {"intkey": 0, "strkey": ""}

    @dataclass
    class DataWithList:
        items: list
        name: str

    data2 = DataWithList(
        items=[{"value": 1, "empty_dict": {}}, {"value": 2, "none_val": None}, [], {}],
        name="test",
    )
    result2 = hlp.ascleandict(data2)
    assert result2 == {"name": "test", "items": [{"value": 1}, {"value": 2}]}


def test_ascleandict_nested_cleanup_multiple_passes():
    """Test that ascleandict removes cascading empty nested structures."""

    @dataclass
    class DeepNested:
        value: str | None = None

    @dataclass
    class MidLevel:
        deep: DeepNested | None = None
        data: dict | None = None

    @dataclass
    class TopLevel:
        mid: MidLevel | None = None
        name: str = "test"

    # Create data where cleaning cascades:
    # - DeepNested.value = None gets removed
    # - MidLevel.deep becomes {} and gets removed
    # - MidLevel.data is already {} and gets removed
    # - MidLevel itself might become {} and gets removed
    data = TopLevel(
        name="test",
        mid=MidLevel(
            deep=DeepNested(value=None),  # Becomes {}
            data={"inner": None},  # Becomes {}
        ),
    )
    result = hlp.ascleandict(data)
    # The while loop should run multiple times, executing line 190
    assert result == {"name": "test"}


def test_unix_timestamp():
    """Test unix_timestamp function."""
    timestamp = hlp.unix_timestamp()
    # Test if it is greater than the timestamp of writing the test
    assert timestamp > 1732266521241
    # and less than 100 years from now
    assert timestamp < 1732266521241 + 100 * 365 * 24 * 3600 * 1000


def test_format_unix_timestamp():
    """Test format_unix_timestamp function."""
    timestamp = hlp.format_unix_timestamp(1732266521241)
    # Test if it is in the correct format
    assert timestamp == "2024-11-22_10-08-41" or timestamp == "2024-11-22_09-08-41"


def test_fw_id_from_fw_file_passes():
    """Test FirmwareID from firmware file."""
    # Test with a valid firmware file
    test_file = "tests/files/fw-test-file.hex"
    res = hlp.fw_id_from_fw_file(test_file, contains="app-nrf9160-wmbus")
    assert "app-nrf9160-wmbus" in res.name


def test_fw_id_from_fw_file_fail():
    """Test FirmwareID from firmware file."""
    # Test with a valid firmware file
    test_file = "tests/files/fw-test-file.hex"
    with pytest.raises(ValueError):
        hlp.fw_id_from_fw_file(test_file, contains="non-existing")
    with pytest.raises(ValueError):
        hlp.fw_id_from_fw_file(test_file)



import pytest

from lob_hlpr import (  # Adjust the import path as necessary
    FirmwareID,
    FirmwareVersion,
)


def test_valid_firmware_id_with_variants():
    """Tests if the firmware ID can be created with variants."""
    id_str = (
        "app-nrf9160-wmbus v0.24.1-9-g8ad003f+hw3.alt_phy TZ2 (Mar 11 2024 13:57:40)"
    )
    firmware_id = FirmwareID(id_str)
    assert firmware_id.name == "app-nrf9160-wmbus"
    assert isinstance(firmware_id.version, FirmwareVersion)
    assert firmware_id.version.version_string == "0.24.1-9-g8ad003f"
    assert firmware_id.variants == ["hw3", "alt_phy"]
    assert firmware_id.built == "2024-03-11T13:57:40"


def test_valid_firmware_id_with_origin_pass():
    """Tests if the firmware ID can be created with 'origin' in the version string."""
    id_str = "app-nrf91-origin+0.3.7+hw4 TZ3 (Mar  5 2025 08:36:26)"
    firmware_id = FirmwareID(id_str)
    assert firmware_id.name == "app-nrf91-origin"
    assert isinstance(firmware_id.version, FirmwareVersion)
    assert firmware_id.version.version_string == "0.3.7"
    assert firmware_id.variants == ["hw4"]
    assert firmware_id.built == "2025-03-05T08:36:26"


def test_valid_firmware_id_unknown():
    """Tests if the firmware ID can be created with 'unknown' in the version string."""
    id_str = "app-nrf9160-wmbus v0.0.0-unknown+hw3 TZ2 (Oct 12 2023 10:45:49)"
    firmware_id = FirmwareID(id_str)
    assert firmware_id.name == "app-nrf9160-wmbus"
    assert isinstance(firmware_id.version, FirmwareVersion)
    assert firmware_id.version.version_string == "0.0.0-unknown"
    assert firmware_id.variants == ["hw3"]
    assert firmware_id.built == "2023-10-12T10:45:49"


def test_valid_firmware_id_without_variants():
    """Tests if the firmware ID can be created without variants."""
    id_str = "Lobaro v1.2.3 (Jan 01 2021 00:00:00)"
    firmware_id = FirmwareID(id_str)
    assert firmware_id.name == "Lobaro"
    assert isinstance(firmware_id.version, FirmwareVersion)
    assert firmware_id.version.version_string == "1.2.3"
    assert firmware_id.variants == []
    assert firmware_id.built == "2021-01-01T00:00:00"


def test_valid_firmware_without_date():
    """Tests if the firmware ID can be created without variants."""
    id_str = "Lobaro v1.2.3"
    firmware_id = FirmwareID(id_str)
    assert firmware_id.name == "Lobaro"
    assert isinstance(firmware_id.version, FirmwareVersion)
    assert firmware_id.version.version_string == "1.2.3"
    assert firmware_id.variants == []
    assert firmware_id.built is None


def test_firmware_id_with_invalid_date():
    """Tests if the firmware ID can be created with an invalid date format."""
    id_str = "Lobaro v1.2.3+hw3 (Invalid Date Format)"
    with pytest.raises(ValueError):
        FirmwareID(id_str)


def test_firmware_id_with_incomplete_string():
    """Tests if the firmware ID can be created with an incomplete string."""
    id_str = "Lobaro v1.2"
    with pytest.raises(ValueError):
        FirmwareID(id_str)


def test_firmware_id_with_no_version():
    """Tests if the firmware ID can be created without a version."""
    id_str = "Lobaro v (Jan 01 2021 00:00:00)"
    with pytest.raises(ValueError):
        FirmwareID(id_str)


def test_firmware_id_with_no_name():
    """Tests if the firmware ID can be created without a name."""
    id_str = " v1.2.3 (Jan 01 2021 00:00:00)"
    with pytest.raises(ValueError):
        FirmwareID(id_str)
